from operator import eq

import networkx as nx
from django.test import TestCase

from core.cytoscape import read_sif, write_sif
from core.models import Node, Edge
from core.network import NetworkService
from core.network_analyser import NetworkAnalyser


class NetworkTestCase(TestCase):
    def setUp(self):
        self.network_service = NetworkService()
        x = Node('x', 0, 2)
        y = Node('y', 0, 1)
        edge1 = Edge(x.id, y.id, 1)
        edge2 = Edge(y.id, x.id, -1)
        edge3 = Edge(x.id, x.id, 2)
        self.network_service.add_node(x)
        self.network_service.add_node(y)
        self.network_service.add_edge(edge1)
        self.network_service.add_edge(edge2)
        self.network_service.add_edge(edge3)

    def test_predecessor_combinations(self):
        predecessor_combinations = NetworkAnalyser.get_predecessors(self.network_service.network)
        expected_x = ((), ('y',), ('x',), ('y', 'x'))
        self.assertTupleEqual(predecessor_combinations['x'], expected_x)

    def test_state_space(self):
        nodes = self.network_service.network.nodes
        result = list(NetworkAnalyser.generate_state_space(self.network_service.network))
        expected = [{'x': 0, 'y': 0}, {'x': 0, 'y': 1}, {'x': 1, 'y': 0}, {'x': 1, 'y': 1}, {'x': 2, 'y': 0},
                    {'x': 2, 'y': 1}]
        self.assertListEqual(result, expected)

    def test_state_graph(self):
        parameters = {
            "x": [
                ([], 0,), (['x'], 2), (['y'], 2), (['x', 'y'], 2)
            ],
            "y": [
                ([], 0,), (['x'], 1)
            ]
        }
        graph = NetworkAnalyser.get_state_graph(self.network_service.network, parameters)
        result = list(graph.nodes)
        expected = ['0,0', '0,1', '1,0', '1,1', '2,0', '2,1']
        self.assertListEqual(result, expected)

    def test_cycles(self):
        network = self.network_service.network
        NetworkAnalyser.get_cycles()

    def test_out_degree(self):
        NetworkAnalyser.get_deadlock_states()

    def test_sif_converter(self):
        write_sif(self.network_service.network, 'test.sif')
        G = read_sif("test.sif")
        em = nx.algorithms.isomorphism.generic_edge_match('weight', 1, lambda x, y: eq(str(x), str(y)))
        result = nx.is_isomorphic(self.network_service.network, G, edge_match=em)
        self.assertTrue(result)

    def tearDown(self):
        self.network_service.clear()



import itertools

from networkx import DiGraph, nx
from networkx.classes.reportviews import EdgeView

from core.pyrthomas import utils


class NetworkAnalyser:

    @staticmethod
    def get_predecessors(network: DiGraph):
        result = dict()
        for node in network.nodes:
            predecessors = tuple(nx.DiGraph.predecessors(network, node))
            combinations = utils.all_subsets(predecessors)
            result[node] = tuple(combinations)
        return result

    @staticmethod
    def generate_state_space(graph: DiGraph):
        nodes = graph.nodes
        parameters = list()
        for node in nodes:
            edges_from_node = graph.edges(node, data=True)
            max_weighted_edge = max(edges_from_node, key=lambda x: utils.get_weight(x, True))
            max_threshold = utils.get_weight(max_weighted_edge, True)
            values = tuple(range(max_threshold + 1))
            parameters.append(values)
        state_space = tuple(itertools.product(*parameters))
        for element in state_space:
            yield dict(zip(nodes, element))

    @staticmethod
    def get_state_graph(network: DiGraph, parameters) -> DiGraph:
        state_space = list(NetworkAnalyser.generate_state_space(network))
        state_graph = nx.DiGraph()
        state_space_nodes = [utils.create_node_from_dict(entry) for entry in state_space]
        state_graph.add_nodes_from(state_space_nodes)

        resources = NetworkAnalyser.calculate_resources(network, state_space)

        k_states = NetworkAnalyser.calculate_k(resources, parameters)

        next_states = NetworkAnalyser.calculate_next_states(k_states)

        edges = NetworkAnalyser.generate_edges(next_states)
        state_graph.add_edges_from(edges)
        return state_graph

    @staticmethod
    def generate_edges(next_states):
        edges = list()
        for key, states in next_states.items():
            previous_value = dict(key)
            source = utils.create_node_from_dict(previous_value)
            for state in states:
                target = utils.create_node_from_dict(state)
                edges.append((source, target))
        return edges

    @staticmethod
    def calculate_next_states(k_states):
        for state_key, state in k_states.items():
            previous_entities = dict(state_key)
            new_val = list()
            temp_previous = previous_entities.copy()
            for entity_key, previous_entity in previous_entities.items():
                next_value = k_states[state_key][entity_key]

                if next_value > previous_entity:
                    previous_entities[entity_key] = previous_entity + 1
                elif next_value < previous_entity:
                    previous_entities[entity_key] = previous_entity - 1

                if not (previous_entities in new_val) and not previous_entities == temp_previous:
                    new_val.append(previous_entities.copy())
                    previous_entities[entity_key] = temp_previous[entity_key]
            k_states[state_key] = new_val
        return k_states

    @staticmethod
    def calculate_k(resources, parameters):
        for resource_key, entities in resources.items():
            for entity_key, entity in entities.items():
                matched_interaction = filter(lambda x: sorted(x[0]) == sorted(entity),
                                             parameters[entity_key])
                first_interaction = next(matched_interaction)
                entities[entity_key] = first_interaction[1]
        return resources

    @staticmethod
    def get_cycles(network: nx.DiGraph):
        return nx.simple_cycles(network)

    @staticmethod
    def get_deadlock_states(network: nx.DiGraph):
        out_degree_iter = network.out_degree(network.nodes)
        return [node for node, out_degree in out_degree_iter if out_degree == 0]

    @staticmethod
    def calculate_resources(network, state_space):
        resources = dict()
        for state in state_space:
            node = dict()
            for key in state:
                entity_resources = list()
                in_edges = network.in_edges(key, data=True)
                for edge in in_edges:
                    is_resource = NetworkAnalyser.is_resource_of_state(state, edge)
                    if is_resource:
                        entity_resources.append(edge[0])
                entity_resources.sort()
                node[key] = entity_resources

            state_key = tuple(state.items())
            resources[state_key] = node
        return resources

    @staticmethod
    def get_required_parameters(network: DiGraph):
        nodes = network.nodes
        parameters = dict()
        for node in nodes:
            predecessors = tuple(nx.DiGraph.predecessors(network, node))
            required_interactions = [{'interaction': interaction, 'value': None} for interaction in
                                     utils.all_subsets(predecessors)]
            parameters[node] = required_interactions
        return parameters

    @staticmethod
    def is_resource_of_state(state: dict, edge: EdgeView):
        weight = utils.get_weight(edge)
        is_positive = weight >= 0
        value = state[edge[0]]
        return (is_positive and value >= abs(weight)) or (not is_positive and value < abs(weight))



import itertools

import networkx as nx
from networkx.classes.reportviews import EdgeView


def all_subsets(items):
    return itertools.chain(*map(lambda x: itertools.combinations(items, x), range(0, len(items) + 1)))


def create_node_from_dict(item: dict):
    return ','.join(map(str, item.values()))


def persist_network(network: nx.Graph, pickle_key):
    nx.write_gpickle(network, pickle_key)


def get_weight(edge: EdgeView, absolute=False):
    weight = edge[2]['weight']
    if absolute:
        return abs(weight)
    return abs(weight) if absolute else weight
